/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdambuffermanagerskin.h"
#include <gtk/gtkclist.h>
#include <gdam/component/gdambuffermanager.h>
#include "gdamskinutil.h"
#include <stdlib.h>
#include <string.h>

static GtkObjectClass* parent_class;

static void gdam_buffer_manager_skin_destroy(GtkObject* object);
static void gdam_buffer_manager_skin_class_init(GdamSkinClass* klass);

static void gdam_buffer_manager_skin_init(GdamBufferManagerSkin* buffer_manager_skin)
{
}

GtkType gdam_buffer_manager_skin_get_type()
{
	static GtkType buffer_manager_skin_type = 0;
	if (!buffer_manager_skin_type) {
		static const GtkTypeInfo buffer_manager_skin_info = {
			"GdamBufferManagerSkin",
			sizeof(GdamBufferManagerSkin),
			sizeof(GdamBufferManagerSkinClass),
			(GtkClassInitFunc) gdam_buffer_manager_skin_class_init,
			(GtkObjectInitFunc) gdam_buffer_manager_skin_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_SKIN;
		parent_class = gtk_type_class(parent);
		buffer_manager_skin_type = gtk_type_unique(parent,
		                                 &buffer_manager_skin_info);
	}
	return buffer_manager_skin_type;
}

typedef struct _CListInfo CListInfo;
typedef char*   (*CListEntryGetter)(GdamBufferManagerInfo*   info);

struct _CListInfo {
	int                        column;
	GdamBufferManagerSkin*       skin;

	/* We reference count the clist. */
	GtkCList*                  clist;
	int			   selected_connect_id;

	int                        num_columns;
	CListEntryGetter*	   column_fillers;

	GdamBufferManagerNotifier* notifier;
};

static void clist_set_default(GtkCList*         clist,
                              int               row,
			      int               column,
			      GdkEvent*         event,
			      CListInfo*        clist_info)
{
	char* buffer_id_txt;
	int buffer_id;
	buffer_id_txt = NULL;
	gtk_clist_get_text(clist, row, clist_info->column, &buffer_id_txt);
	if (!buffer_id_txt
	 || (buffer_id=atoi(buffer_id_txt)) <= 0) {
	 	g_warning("error getting buffer id from %d,%d in clist",
				row, clist_info->column);
		return;
	}
	gdam_buffer_set_default(buffer_id);
}

static void clist_buffer_add_callback(GdamBufferManagerInfo*   info,
                                      gpointer                 user_data)
{
	CListInfo* clist_info = (CListInfo*) user_data;
	GtkCList* clist = clist_info->clist;
	char** strings;
	int i;
	int num_columns = clist_info->num_columns;
	int new_idx;
	g_return_if_fail(num_columns == clist->columns);
	strings = (char**)alloca(sizeof(char*) * num_columns);
	for (i = 0; i < num_columns; i++) {
		strings[i] = (*clist_info->column_fillers[i])(info);
		gdam_debug("buffer_add_callback: strings[%d]=%s",
			i, strings[i]);
	}
	new_idx = gtk_clist_append(clist, strings);
	/* TODO: is this safe?   what if info disappears? */
	gtk_clist_set_row_data(clist, new_idx, info);
	for (i = 0; i < num_columns; i++)
		g_free(strings[i]);
}

static void clist_buffer_remove_callback(GdamBufferManagerInfo*   info,
                                         gpointer                 user_data)
{
	CListInfo* clist_info = (CListInfo*) user_data;
	GtkCList* clist = clist_info->clist;
	int i;
	for (i = 0; i < clist->rows; i++) {
		char* txt = NULL;
		int id;
		gtk_clist_get_text(clist, i, clist_info->column, &txt);
		if (txt == NULL || (id=atoi(txt)) <= 0)
			continue;
		if (id == info->buffer_id) {
			gtk_clist_remove(clist, i);
			return;
		}
	}
}

static gboolean is_one_of(const char* key, ...)
{
	va_list args;
	char* cmp;
	va_start(args, key);
	while ((cmp = va_arg(args, char*)) != NULL) {
		if (g_strcasecmp(key, cmp) == 0) {
			va_end(args);
			return TRUE;
		}
	}
	va_end(args);
	return FALSE;
}

static int find_clist_column(int num_cols, char** cols)
{
	int out = 0;
	while (out < num_cols && *cols != NULL) {
		if (is_one_of(*cols++,
		              "buffer id", "buffer-id", "buffer_id",
		              (char*) NULL))
		 	return out;
		out++;
	}
	return -1;
}
static char* buffer_id_getter(GdamBufferManagerInfo* info)
{
	return g_strdup_printf("%d", info->buffer_id);
}
static char* description_getter(GdamBufferManagerInfo* info)
{
	return g_strdup(info->description);
}
static char* unknown_getter(GdamBufferManagerInfo* info)
{
	return g_strdup("unknown");
}
static CListEntryGetter get_column_filler(const char* title)
{
	if (is_one_of(title, "buffer_id",
	                     "buffer-id",
			     "buffer id", (char*) NULL))
		return buffer_id_getter;

	if (is_one_of(title, "description", (char*) NULL))
		return description_getter;

	return unknown_getter;
}

static void on_scrlist_buffer_drag(GtkSelectionData* selection_data,
				   gpointer	     udata)
{
	CListInfo* info = (CListInfo*)udata;
	GtkCList* clist = info->clist;
	GList* selected = clist->selection;
	int num = GPOINTER_TO_INT(selected->data);
	gpointer data = gtk_clist_get_row_data(clist, num);
	GdamBufferManagerInfo* buf_info = (GdamBufferManagerInfo*)data;
	char str[32];
	sprintf(str, "%d", buf_info->buffer_id);

	gtk_selection_data_set (selection_data, selection_data->target, 8,
				(guchar*)strdup(str), strlen(str));
}

static void connect_clist(GdamBufferManagerSkin* skin,
                          char**                 column_names,
			  GtkCList*              clist)
{
	CListInfo* info = g_new(CListInfo, 1);
	int i;
	info->selected_connect_id
			= gtk_signal_connect(GTK_OBJECT(clist),
					     "select_row",
			                     GTK_SIGNAL_FUNC(clist_set_default),
			                     info);
	gdam_widget_string_source_set(GTK_WIDGET(clist),
				      on_scrlist_buffer_drag,
				      "gdam-buffer-id-text",
				      info);
				      
	info->skin = skin;
	info->clist = clist;
	info->num_columns = clist->columns;
	info->column = find_clist_column(info->num_columns, column_names);
	if (info->column < 0)
		g_warning("no clist column titled Buffer Id");
	info->column_fillers = g_new(CListEntryGetter, info->num_columns);
	for (i = 0; i < info->num_columns; i++) {
		char* title = column_names[i];
		if (title == NULL)
			break;
		gdam_debug("filler for column %d based on %s", i, title);
		info->column_fillers[i] = get_column_filler(title);
	}

	for (  ; i < info->num_columns; i++)
		info->column_fillers[i] = get_column_filler(NULL);

	gtk_object_ref(GTK_OBJECT(clist));
	skin->clist_list = g_slist_prepend(skin->clist_list, info);

	info->notifier = g_new(GdamBufferManagerNotifier, 1);
	info->notifier->buffer_add_callback = clist_buffer_add_callback;
	info->notifier->buffer_remove_callback = clist_buffer_remove_callback;
	info->notifier->user_data = info;
	gdam_buffer_manager_subscribe(info->notifier);
}

static void disconnect_clist(gpointer value, gpointer user_data)
{
	CListInfo* info = (CListInfo*)value;
	gtk_signal_disconnect(GTK_OBJECT(info->clist),
	                      info->selected_connect_id);
	gdam_buffer_manager_unsubscribe(info->notifier);
	g_free(info);
}

static gboolean gdam_buffer_manager_clist_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	if (!GTK_IS_CLIST(object)) {
		g_warning("%s only valid for CList's", handler_names[0]);
		return TRUE;
	}
	connect_clist(GDAM_BUFFER_MANAGER_SKIN(skin),
	              handler_names + 1,
		      GTK_CLIST(object));
	return TRUE;
}

static void do_remove_buffer(GdamBufferManagerSkin* skin)
{
	gdam_buffer_manager_soft_destroy (gdam_buffer_get_default());
}
static gboolean gdam_buffer_manager_remove_selected_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(object,
	                          signal_name,
				  GTK_SIGNAL_FUNC(do_remove_buffer),
				  (GtkObject*) skin);
	return TRUE;
}

static void do_destroy_buffer(GdamBufferManagerSkin* skin)
{
	/** XXX: we use the default buffer, not the buffer selected
	         here! (in the clists...) */
	gdam_buffer_manager_hard_destroy (gdam_buffer_get_default());
}

static gboolean gdam_buffer_manager_destroy_selected_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(object,
	                          signal_name,
				  GTK_SIGNAL_FUNC(do_destroy_buffer),
				  (GtkObject*) skin);
	return TRUE;
}

static void gdam_buffer_manager_skin_destroy(GtkObject* object)
{
	GdamBufferManagerSkin* skin = GDAM_BUFFER_MANAGER_SKIN(object);
	g_slist_foreach(skin->clist_list, disconnect_clist, NULL);
	g_slist_free(skin->clist_list);
	skin->clist_list = NULL;
	(*parent_class->destroy)(object);
}

static GdamSkinGenericConnect connect_infos[] =
{ 	{ "gdam_buffer_manager_clist",
	   gdam_buffer_manager_clist_handler },
        { "gdam_buffer_manager_remove_selected",
	   gdam_buffer_manager_remove_selected_handler },
        { "gdam_buffer_manager_destroy_selected",
	   gdam_buffer_manager_destroy_selected_handler },
	
};

static void gdam_buffer_manager_skin_class_init(GdamSkinClass* klass)
{
	GdamSkinClass* skin_class = GDAM_SKIN_CLASS(klass);
	GTK_OBJECT_CLASS(klass)->destroy = gdam_buffer_manager_skin_destroy;
	gdam_skin_class_register_generic_handlers
	                                      (skin_class,
				               G_N_ELEMENTS(connect_infos),
				               connect_infos);
}

